package geniusweb.voting;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotEquals;
import static org.junit.Assert.assertTrue;
import static org.mockito.Mockito.mock;

import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.junit.Before;
import org.junit.Test;

import geniusweb.actions.PartyId;
import geniusweb.actions.Vote;
import geniusweb.actions.Votes;
import geniusweb.issuevalue.Bid;
import tudelft.utilities.junit.GeneralTests;

public class CollectedVotesTest extends GeneralTests<CollectedVotes> {

	private final Bid bidA = mock(Bid.class), bidB = mock(Bid.class);
	private final Bid a = mock(Bid.class), b = mock(Bid.class),
			c = mock(Bid.class), d = mock(Bid.class);
	private final PartyId party1 = new PartyId("party1");
	private final PartyId party2 = new PartyId("party2");
	private final PartyId party3 = new PartyId("party3");
	private final PartyId party4 = new PartyId("party4");

	private final PartyId partyP = new PartyId("partyP");
	private final PartyId partyQ = new PartyId("partyQ");
	private final PartyId partyR = new PartyId("partyR");
	private final PartyId partyS = new PartyId("partyS");

	private final Vote votePA2 = new Vote(partyP, bidA, 2, 9);
	private final Vote voteQA2 = new Vote(partyQ, bidA, 2, 9);
	private final Vote voteQA3 = new Vote(partyQ, bidA, 3, 9);
	private final Vote voteQB3 = new Vote(partyQ, bidB, 3, 9);
	private final Vote voteQB2 = new Vote(partyQ, bidB, 2, 9);
	private final Vote voteRA2 = new Vote(partyR, bidA, 2, 9);
	private final Vote voteRA3 = new Vote(partyR, bidA, 3, 9);
	private final Vote voteRA4 = new Vote(partyR, bidA, 4, 9);
	private final Vote voteRB2 = new Vote(partyR, bidB, 2, 9);

	private Map<PartyId, Integer> allpowers = new HashMap<>();

	private CollectedVotes cv1 = new CollectedVotes(Collections.emptyMap(),
			Collections.emptyMap());
	private CollectedVotes cv1a = new CollectedVotes(Collections.emptyMap(),
			Collections.emptyMap());

	private Votes votes = new Votes(partyP, Collections.singleton(votePA2));
	private CollectedVotes cv2 = new CollectedVotes(
			Collections.singletonMap(partyP, votes),
			Collections.singletonMap(partyP, 1));

	@Override
	public List<List<CollectedVotes>> getGeneralTestData() {
		return Arrays.asList(Arrays.asList(cv1, cv1a), Arrays.asList(cv2));
	}

	@Override
	public List<String> getGeneralTestStrings() {
		return Arrays.asList("CollectedVotes.\\{\\}.*\\{\\}.*",
				"CollectedVotes.\\{partyP=Votes.*Vote.partyP,Mock.*,2.*\\}.*\\{partyP=1\\}.*");
	}

	@Before
	public void before() {
		allpowers.put(partyP, 1);
		allpowers.put(partyQ, 1);
		allpowers.put(partyR, 1);
		allpowers.put(partyS, 1);
	}

	@Test
	public void testGetAllBids() {
		Map<PartyId, Votes> allvotes = new HashMap<>();
		allvotes.put(partyP, new Votes(partyP, Collections.singleton(votePA2)));
		allvotes.put(partyQ, new Votes(partyQ,
				new HashSet<>(Arrays.asList(voteQA2, voteQB3))));

		Map<Bid, Set<Vote>> allbids = new CollectedVotes(allvotes, allpowers)
				.getAllBids();
		assertEquals(2, allbids.size());
		assertEquals(new HashSet<>(Arrays.asList(votePA2, voteQA2)),
				allbids.get(bidA));
		assertEquals(new HashSet<>(Arrays.asList(voteQB3)), allbids.get(bidB));
	}

	@Test
	public void testGetMaxSubset2() {
		Set<Vote> votes = new HashSet<>(Arrays.asList(votePA2, voteQA2));
		CollectedVotes cv = new CollectedVotes(Collections.emptyMap(),
				allpowers);
		Set<PartyId> parties = cv.getMaxPowerGroup(votes);
		assertEquals(new HashSet<>(Arrays.asList(partyP, partyQ)), parties);

		// test the optimized code similarly
		parties = cv.getMaxPowerGroupBreathFirst(votes);
		assertEquals(new HashSet<>(Arrays.asList(partyP, partyQ)), parties);

	}

	@Test
	public void testGetMaxSubset3() {
		Set<Vote> votes = new HashSet<>(
				Arrays.asList(votePA2, voteQA2, voteRA4));
		CollectedVotes cv = new CollectedVotes(Collections.emptyMap(),
				allpowers);

		Set<PartyId> parties = cv.getMaxPowerGroup(votes);
		// party R wants 4 votes so we can't satisfy
		assertEquals(new HashSet<>(Arrays.asList(partyP, partyQ)), parties);

		// test the optimized code similarly
		parties = cv.getMaxPowerGroupBreathFirst(votes);
		assertEquals(new HashSet<>(Arrays.asList(partyP, partyQ)), parties);
	}

	@Test
	public void testGetMaxSubset3b() {
		Set<Vote> votes = new HashSet<>(
				Arrays.asList(votePA2, voteQA2, voteRA3));
		CollectedVotes cv = new CollectedVotes(Collections.emptyMap(),
				allpowers);
		Set<PartyId> parties = cv.getMaxPowerGroup(votes);
		// party R wants 4 votes so we can't satisfy
		assertEquals(new HashSet<>(Arrays.asList(partyP, partyQ, partyR)),
				parties);

		// test the optimized code similarly
		parties = cv.getMaxPowerGroupBreathFirst(votes);
		assertEquals(new HashSet<>(Arrays.asList(partyP, partyQ, partyR)),
				parties);
	}

	@Test
	public void testNoConsensus() {
		Set<Vote> votes = new HashSet<>(Arrays.asList(voteQA3, voteRA3));
		CollectedVotes cv = new CollectedVotes(Collections.emptyMap(),
				allpowers);
		assertEquals(Collections.emptySet(), cv.getMaxPowerGroup(votes));
		assertEquals(Collections.emptySet(),
				cv.getMaxPowerGroupBreathFirst(votes));
	}

	@Test
	public void testGetMaxAgreements() {
		Map<PartyId, Votes> allvotes = new HashMap<>();
		allvotes.put(partyP, new Votes(partyP, Collections.singleton(votePA2)));
		allvotes.put(partyQ, new Votes(partyQ,
				new HashSet<>(Arrays.asList(voteQA2, voteQB2))));
		allvotes.put(partyR, new Votes(partyR,
				new HashSet<>(Arrays.asList(voteRA4, voteRB2))));

		Map<Bid, Set<PartyId>> allagrees = new CollectedVotes(allvotes,
				allpowers).getMaxAgreements();
		assertEquals(new HashSet<>(Arrays.asList(partyP, partyQ)),
				allagrees.get(bidA));
		assertEquals(new HashSet<>(Arrays.asList(partyQ, partyR)),
				allagrees.get(bidB));

	}

	@Test
	public void testGetNoAgreements() {
		Map<PartyId, Votes> allvotes = new HashMap<>();
		allvotes.put(partyP, new Votes(partyP, Collections.singleton(votePA2)));
		Map<Bid, Set<PartyId>> allagrees = new CollectedVotes(allvotes,
				allpowers).getMaxAgreements();
		assertEquals(Collections.emptyMap(), allagrees);

	}

	@Test
	public void getConsensusGroupsTest() {
		CollectedVotes cv = new CollectedVotes(Collections.emptyMap(),
				allpowers);

		Set<Set<PartyId>> consensus = cv.getConsensusGroups(
				new HashSet<>(Arrays.asList(votePA2, voteQA2)));
		// should give 1 solution: both parties join in.
		System.out.println(consensus);
		assertEquals(1, consensus.size());
		assertEquals(2, consensus.iterator().next().size());
	}

	@Test
	public void getConsensusGroupsTestPQR() {
		CollectedVotes cv = new CollectedVotes(Collections.emptyMap(),
				allpowers);

		Set<Set<PartyId>> consensus = cv.getConsensusGroups(
				new HashSet<>(Arrays.asList(votePA2, voteQA2, voteRA4)));
		// should give 1 solution: P and Q join in. 4 is unreachable.
		System.out.println(consensus);
		assertEquals(1, consensus.size());
		assertEquals(2, consensus.iterator().next().size());
	}

	@Test
	public void getConsensusGroupsTestPQR2() {
		CollectedVotes cv = new CollectedVotes(Collections.emptyMap(),
				allpowers);

		Set<Set<PartyId>> consensus = cv.getConsensusGroups(
				new HashSet<>(Arrays.asList(votePA2, voteQA2, voteRA2)));
		// should give 4 solutions: P+Q, P+R, Q+R and P+Q+R
		System.out.println(consensus);
		assertEquals(4, consensus.size());
	}

	@Test
	public void getConsensusGroupsTestRExtraPower() {
		// give R extra power to reach the 4 if other 2 join
		allpowers = new HashMap<>();
		allpowers.put(partyP, 1);
		allpowers.put(partyQ, 1);
		allpowers.put(partyR, 2);

		CollectedVotes cv = new CollectedVotes(Collections.emptyMap(),
				allpowers);

		Set<Set<PartyId>> consensus = cv.getConsensusGroups(
				new HashSet<>(Arrays.asList(votePA2, voteQA2, voteRA4)));
		// should give solutions P+Q and P+Q+R
		System.out.println(consensus);
		assertEquals(2, consensus.size());
		Iterator<Set<PartyId>> it = consensus.iterator();
		assertEquals(2, it.next().size());
		assertEquals(3, it.next().size());
	}

	@Test
	public void testWith() {
		Map<PartyId, Votes> allvotes = new HashMap<>();
		allvotes.put(partyP, new Votes(partyP, Collections.singleton(votePA2)));
		CollectedVotes cv = new CollectedVotes(allvotes, allpowers);
		assertEquals(1, cv.getVotes().keySet().size());
		cv = cv.with(new Votes(partyQ, Collections.singleton(voteQA2)), 2);
		assertEquals(2, cv.getVotes().keySet().size());
	}

	@Test
	public void testWithout() {
		Map<PartyId, Votes> allvotes = new HashMap<>();
		allvotes.put(partyP, new Votes(partyP, Collections.singleton(votePA2)));
		allvotes.put(partyQ, new Votes(partyQ,
				new HashSet<>(Arrays.asList(voteQA2, voteQB2))));
		allvotes.put(partyR, new Votes(partyR,
				new HashSet<>(Arrays.asList(voteRA4, voteRB2))));

		// give R extra power to reach the 4 if other 2 join
		allpowers = new HashMap<>();
		allpowers.put(partyP, 1);
		allpowers.put(partyQ, 1);
		allpowers.put(partyR, 2);

		CollectedVotes cv = new CollectedVotes(allvotes, allpowers);

		assertEquals(3, cv.getVotes().size());
		cv = cv.without(new HashSet<>(Arrays.asList(partyP, partyQ)));
		assertEquals(1, cv.getVotes().size());
		assertEquals(1, cv.getPowers().size());

	}

	@Test
	public void getConsensusGroupMaxPower() {
		Vote P = new Vote(partyP, bidA, 2, 3);
		Vote Q = new Vote(partyQ, bidA, 2, 3);
		Vote R = new Vote(partyR, bidA, 2, 3);
		Vote S = new Vote(partyS, bidA, 2, 3);
		Map<PartyId, Votes> allvotes = new HashMap<>();
		allvotes.put(partyP, new Votes(partyP, Collections.singleton(P)));
		allvotes.put(partyQ, new Votes(partyQ, Collections.singleton(Q)));
		allvotes.put(partyR, new Votes(partyR, Collections.singleton(R)));
		allvotes.put(partyS, new Votes(partyS, Collections.singleton(S)));

		CollectedVotes cv = new CollectedVotes(allvotes, allpowers);
		Set<Set<PartyId>> consensus = cv.getConsensusGroups(
				new HashSet<Vote>(Arrays.asList(P, Q, R, S)));
		System.out.println(consensus);
		// check that all are size 2 or 3.
		for (Set<PartyId> set : consensus) {
			assertTrue(set.size() == 2 || set.size() == 3);
		}

	}

	@Test
	public void getConsensusGroupMaxPowerModifiedPartyPowers() {
		// party R gets power 2.
		Vote P = new Vote(partyP, bidA, 2, 3);
		Vote Q = new Vote(partyQ, bidA, 2, 3);
		Vote R = new Vote(partyR, bidA, 2, 3);
		Map<PartyId, Votes> allvotes = new HashMap<>();
		allvotes.put(partyP, new Votes(partyP, Collections.singleton(P)));
		allvotes.put(partyQ, new Votes(partyQ, Collections.singleton(Q)));
		allvotes.put(partyR, new Votes(partyR, Collections.singleton(R)));

		Map<PartyId, Integer> powers = new HashMap<PartyId, Integer>();
		powers.put(partyP, 1);
		powers.put(partyQ, 1);
		powers.put(partyR, 2);

		CollectedVotes cv = new CollectedVotes(allvotes, powers);
		Set<Set<PartyId>> consensus = cv
				.getConsensusGroups(new HashSet<Vote>(Arrays.asList(P, Q, R)));
		System.out.println(consensus);
		// R has power 2 now. That means groups of size 3 are not posisble.
		// check that all are size 2.
		for (Set<PartyId> set : consensus) {
			assertTrue(set.size() == 2);
		}

	}

	@Test
	public void testRandomOrderOfSet() {
		// this test that a Set of parties does not have
		// the fixed order of the list #1878
		// not really testing our code but crucial for proper working.
		List<PartyId> list = Arrays.asList(partyQ, partyR, partyS);
		// #1878 does Set randomize somewhat?
		Set<PartyId> parties = new HashSet<>(list);
		assertNotEquals(list, new LinkedList<>(parties));

		LinkedList<PartyId> list2 = new LinkedList<>(list);
		list2.add(partyP);
		parties = new HashSet<>(list2);
		assertNotEquals(list2, new LinkedList<>(parties));

	}
}
